from enum import Enum
from .time_frame import TimeFrame


class TimeInterval(Enum):
    CURRENT = "CURRENT"
    MINUTES_ONE = "MINUTES_ONE"
    MINUTES_FIFTEEN = "MINUTES_FIFTEEN"
    HOURS_ONE = "HOURS_ONE"
    HOURS_FOUR = "HOURS_FOUR"
    DAYS_ONE = "DAYS_ONE"

    @staticmethod
    def from_string(value: str):

        if isinstance(value, str):

            for entry in TimeInterval:

                if value.upper() == entry.value:
                    return entry

            raise ValueError(
                f"Could not convert {value} to TimeInterval"
            )

    def equals(self, other):

        if isinstance(other, Enum):
            return self.value == other.value

        else:
            return TimeInterval.from_string(other) == self

    @staticmethod
    def from_time_frame(time_frame):

        if TimeFrame.CURRENT.equals(time_frame):
            return TimeInterval.CURRENT
        elif TimeFrame.ONE_HOUR.equals(time_frame):
            return TimeInterval.MINUTES_ONE
        elif TimeFrame.ONE_DAY.equals(time_frame):
            return TimeInterval.MINUTES_FIFTEEN
        elif TimeFrame.ONE_WEEK.equals(time_frame):
            return TimeInterval.HOURS_ONE
        elif TimeFrame.ONE_MONTH.equals(time_frame):
            return TimeInterval.HOURS_FOUR
        elif TimeFrame.ONE_YEAR.equals(time_frame):
            return TimeInterval.DAYS_ONE
        else:
            raise NotImplementedError(
                f"Timeframe {time_frame} not implemented"
            )

    def amount_of_data_points(self):

        if TimeInterval.CURRENT.equals(self):
            return 1
        elif TimeInterval.MINUTES_ONE.equals(self):
            return 60
        elif TimeInterval.MINUTES_FIFTEEN.equals(self):
            return 96
        elif TimeInterval.HOURS_ONE.equals(self):
            return 168
        elif TimeInterval.HOURS_FOUR.equals(self):
            return 168
        elif TimeInterval.DAYS_ONE.equals(self):
            return 365
        else:
            raise NotImplementedError(f"Timeframe {self} not implemented")

    @property
    def time_frame(self):
        from investing_algorithm_framework.core.models.time_frame import \
            TimeFrame

        if TimeInterval.MINUTES_ONE.equals(self):
            return TimeFrame.ONE_HOUR
        elif TimeInterval.MINUTES_FIFTEEN.equals(self):
            return TimeFrame.ONE_DAY
        elif TimeInterval.HOURS_ONE.equals(self):
            return TimeFrame.ONE_WEEK
        elif TimeInterval.HOURS_FOUR.equals(self):
            return TimeFrame.ONE_MONTH
        elif TimeInterval.DAYS_ONE.equals(self):
            return TimeFrame.ONE_YEAR
        else:
            raise NotImplementedError(
                f"TimeInterval {self.value} not implemented"
            )



from abc import ABC, abstractmethod

from investing_algorithm_framework.core.context.algorithm_context import \
    AlgorithmContext
from investing_algorithm_framework.core.context \
    .algorithm_context_configuration import AlgorithmContextConfiguration


class AlgorithmContextInitializer(ABC):

    from investing_algorithm_framework.core.context.algorithm_context \
        import AlgorithmContext

    @abstractmethod
    def initialize(self, algorithm: AlgorithmContext) -> None:
        raise NotImplementedError()


algorithm = AlgorithmContext()

__all__ = [
    'algorithm',
    'AlgorithmContext',
    'AlgorithmContextInitializer',
    'AlgorithmContextConfiguration'
]



from investing_algorithm_framework.core.mixins import BinanceDataProviderMixin
from investing_algorithm_framework.core.data_providers.data_provider import \
    DataProvider


class BinanceDataProvider(BinanceDataProviderMixin, DataProvider):
    pass



import click

from investing_algorithm_framework import current_app as app


@click.group()
def cli():
    pass


@cli.command()
def start():
    app.start_algorithm()


@cli.command()
def stop():
    app.stop_algorithm()


@cli.command()
def orders():
    print(app.algorithm.get_orders())


@cli.command()
def positions():
    print(app.algorithm.get_positions())


@cli.command()
def portfolio():
    print(app.algorithm.get_portfolio())


if __name__ == '__main__':
    cli()



import os

from enum import Enum
import logging

from investing_algorithm_framework.core import OperationalException
from investing_algorithm_framework.configuration.constants import \
    DATABASE_NAME, DATABASE_DIRECTORY_PATH, RESOURCES_DIRECTORY

logger = logging.getLogger(__name__)


class Environment(Enum):
    """
    Class TimeUnit: Enum for data_provider base type
    """
    DEV = 'DEV'
    PROD = 'PROD'
    TEST = 'TEST'

    # Static factory method to convert
    # a string to environment
    @staticmethod
    def from_string(value: str):

        if isinstance(value, str):

            for entry in Environment:

                if value.upper() == entry.value:
                    return entry

            raise OperationalException(
                "Could not convert value to a environment type"
            )

        else:
            raise OperationalException(
                "Could not convert non string value to a environment type"
            )

    def equals(self, other):

        if isinstance(other, Enum):
            return self.value == other.value
        else:

            try:
                data_base_type = Environment.from_string(other)
                return data_base_type == self
            except OperationalException:
                pass

            return other == self.value


class Config(dict):
    ENV = "DEV"
    LOG_LEVEL = 'DEBUG'
    APP_DIR = os.path.abspath(os.path.dirname(__file__))
    PROJECT_ROOT = os.path.abspath(os.path.join(APP_DIR, os.pardir))
    DEBUG_TB_INTERCEPT_REDIRECTS = False
    SQLALCHEMY_TRACK_MODIFICATIONS = False
    CACHE_TYPE = 'simple'  # Can be "memcached", "redis", etc.
    CORS_ORIGIN_WHITELIST = [
        'http://0.0.0.0:4100',
        'http://localhost:4100',
        'http://0.0.0.0:8000',
        'http://localhost:8000',
        'http://0.0.0.0:4200',
        'http://localhost:4200',
        'http://0.0.0.0:4000',
        'http://localhost:4000',
    ]
    RESOURCES_DIRECTORY = os.getenv(RESOURCES_DIRECTORY)
    DATABASE_CONFIG = {
        DATABASE_NAME: os.getenv(DATABASE_NAME, "database"),
        DATABASE_DIRECTORY_PATH: os.getenv(DATABASE_DIRECTORY_PATH)
    }
    SCHEDULER_API_ENABLED = True
    CHECK_PENDING_ORDERS = True

    def __init__(self):
        super().__init__()

        for attribute_key in dir(Config):

            if attribute_key.isupper():
                self[attribute_key] = getattr(self, attribute_key)

    def __setitem__(self, key, item):
        self.__dict__[key] = item

    def __getitem__(self, key):
        return self.__dict__[key]

    def __repr__(self):
        return repr(self.__dict__)

    def __len__(self):
        return len(self.__dict__)

    def __delitem__(self, key):
        del self.__dict__[key]

    def clear(self):
        return self.__dict__.clear()

    def copy(self):
        return self.__dict__.copy()

    def has_key(self, k):
        return k in self.__dict__

    def update(self, *args, **kwargs):
        return self.__dict__.update(*args, **kwargs)

    def keys(self):
        return self.__dict__.keys()

    def values(self):
        return self.__dict__.values()

    def items(self):
        return self.__dict__.items()

    def pop(self, *args):
        return self.__dict__.pop(*args)

    def __cmp__(self, dict_):
        return self.__cmp__(self.__dict__, dict_)

    def __contains__(self, item):
        return item in self.__dict__

    def __iter__(self):
        return iter(self.__dict__)

    def __str__(self):
        field_strings = []

        for attribute_key in self:

            if attribute_key.isupper():
                field_strings.append(
                    f'{attribute_key}='
                    f'{self[attribute_key]!r}'
                )

        return f"<{self.__class__.__name__}({','.join(field_strings)})>"

    def get(self, key: str, default=None):
        """
        Mimics the dict get() functionality
        """

        try:
            return self[key]
        # Ignore exception
        except Exception:
            pass

        return default

    def set(self, key: str, value) -> None:
        self[key] = value

    @staticmethod
    def from_dict(dictionary):
        config = Config()

        for attribute_key in dictionary:

            if attribute_key:
                config.set(attribute_key, dictionary[attribute_key])
                config[attribute_key] = dictionary[attribute_key]
        return config


class TestConfig(Config):
    ENV = Environment.TEST.value
    TESTING = True
    DATABASE_CONFIG = {
        'DATABASE_NAME': "test",
    }


class DevConfig(Config):
    ENV = Environment.DEV.value
    DATABASE_CONFIG = {
        'DATABASE_NAME': "dev",
    }



import setuptools
from investing_algorithm_framework import get_version

VERSION = get_version()

with open("README.md", "r") as fh:
    long_description = fh.read()

with open('requirements.txt') as f:
    required = f.read().splitlines()

setuptools.setup(
    name="investing_algorithm_framework",
    version=get_version(),
    license='Apache License 2.0',
    author="coding kitties",
    description="A framework for creating an investment algorithm",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/coding-kitties/investing-algorithm-framework.git",
    download_url='https://github.com/coding-kitties/investing-algorithm-framework/archive/v0.1.1.tar.gz',
    packages=setuptools.find_packages(exclude=['tests', 'tests.*']),
    keywords=['INVESTING', 'BOT', 'ALGORITHM', 'FRAMEWORK'],
    classifiers=[
        "Intended Audience :: Developers",
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        "Topic :: Software Development",
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
    ],
    install_requires=required,
    python_requires='>=3',
    include_package_data=True,
)



import csv
import shutil
import tempfile
from typing import Dict, List

from investing_algorithm_framework.core.exceptions import OperationalException


# Function to add column headers to csv
def add_column_headers_to_csv(file_name: str, column_names: List[str]) -> None:

    # Open file in append mode
    with open(file_name, 'a+', newline='') as write_obj:

        # Create a writer object from csv module
        csv_writer = csv.writer(write_obj)

        # Add the columns
        csv_writer.writerow(column_names)


# Function to append dict as a row to csv
def append_dict_as_row_to_csv(
        file_name: str, dict_data: Dict, field_names: List[str]
) -> None:
    result_dict = {field: dict_data[field] for field in field_names}

    # Open file in append mode
    with open(file_name, 'a', newline='') as write_obj:
        # Create a writer object from csv module
        dict_writer = csv.DictWriter(write_obj, fieldnames=field_names)
        # Add dictionary as wor in the csv
        dict_writer.writerow(result_dict)


def get_total_amount_of_rows(file_path: str):
    file = open(file_path)
    reader = csv.reader(file)
    rows_count = len(list(reader))
    file.close()
    return rows_count


# Function to convert a csv to a list of data_provider
def csv_to_list(csv_path: str, strip_column_headers: bool = False) -> List:
    data = []
    with open(csv_path, "r") as csv_f:
        reader = csv.reader(csv_f, delimiter=',')

        for row in reader:

            if row:  # avoid blank lines
                data.append(row)

    # Delete columns
    if strip_column_headers:
        del data[0]

    return data


# Function to remove a row of a csv corresponding to a particular index
def remove_row(csv_path: str, row_index: int) -> None:

    if row_index < 0:
        raise OperationalException('Negative index number')

    _, temp_file = tempfile.mkstemp(suffix='.csv')

    with open(csv_path, 'r') as inf, open(temp_file, 'w') as out_f:
        writer = csv.writer(out_f)
        index = 0

        for row in csv.reader(inf):

            # Copy every row except the index row
            if index != row_index:
                writer.writerow(row)

            index += 1

    shutil.move(temp_file, csv_path)



from abc import ABC, abstractmethod
from investing_algorithm_framework.core.exceptions import OperationalException
from investing_algorithm_framework.core.models import OrderSide, OrderType


class OrderValidator(ABC):

    def validate(self, order, portfolio):

        if OrderSide.BUY.equals(order.order_side):
            self.validate_buy_order(order, portfolio)
        else:
            self.validate_sell_order(order, portfolio)

        if OrderType.LIMIT.equals(order.order_type):
            self.validate_limit_order(order, portfolio)
        else:
            self.validate_market_order(order, portfolio)

        self._validate_order(order, portfolio)

    def validate_sell_order(self, order, portfolio):

        position = portfolio.positions\
            .filter_by(portfolio=portfolio)\
            .filter_by(symbol=order.target_symbol)\
            .first()

        if position is None:
            raise OperationalException(
                "Can't add sell order to non existing position"
            )

        if OrderType.LIMIT.equals(order.order_type):
            if position.amount < order.amount_target_symbol:
                raise OperationalException(
                    "Order amount is larger then amount of open position"
                )
        else:
            if position.amount < order.amount_target_symbol:
                raise OperationalException(
                    "Order amount is larger then amount of open position"
                )

        if not order.trading_symbol == portfolio.trading_symbol:
            raise OperationalException(
                f"Can't add sell order with target "
                f"symbol {order.target_symbol} to "
                f"portfolio with trading currency {portfolio.trading_symbol}"
            )

    @staticmethod
    def validate_buy_order(order, portfolio):

        if not order.trading_symbol == portfolio.trading_symbol:
            raise OperationalException(
                f"Can't add buy order with trading "
                f"symbol {order.trading_symbol} to "
                f"portfolio with trading currency {portfolio.trading_symbol}"
            )

    @staticmethod
    def validate_limit_order(order, portfolio):

        total_price = order.amount_target_symbol * order.initial_price

        if float(portfolio.unallocated) < total_price:
            raise OperationalException(
                f"Order total: {total_price} {portfolio.trading_symbol}, is "
                f"larger then unallocated size: {portfolio.unallocated} "
                f"{portfolio.trading_symbol} of the portfolio"
            )

    @staticmethod
    def validate_market_order(order, portfolio):

        if OrderSide.BUY.equals(order.order_side):

            if order.amount_trading_symbol is None:
                raise OperationalException(
                    f"Market order needs an amount specified in the trading "
                    f"symbol {order.trading_symbol}"
                )

            if order.amount_trading_symbol > portfolio.unallocated:
                raise OperationalException(
                    f"Market order amount {order.amount_trading_symbol} "
                    f"{portfolio.trading_symbol.upper()} is larger then "
                    f"unallocated {portfolio.unallocated} "
                    f"{portfolio.trading_symbol.upper()}"
                )
        else:
            position = portfolio.positions\
                .filter_by(symbol=order.target_symbol)\
                .first()

            if position is None:
                raise OperationalException(
                    "Can't add market sell order to non existing position"
                )

            if order.amount_target_symbol > position.amount:
                raise OperationalException(
                    "Sell order amount larger then position size"
                )

    @abstractmethod
    def _validate_order(self, order, portfolio):
        pass



import ccxt
import logging
from investing_algorithm_framework.core.market_services.market_service \
    import MarketService
from investing_algorithm_framework.core.exceptions import OperationalException
from investing_algorithm_framework.core.mixins import \
    BinanceApiSecretKeySpecifierMixin
from investing_algorithm_framework.configuration.constants import BINANCE, \
    BINANCE_API_KEY, BINANCE_SECRET_KEY
from investing_algorithm_framework.core.models import TimeInterval


BINANCE_CCXT_ID = "binance"
logger = logging.getLogger(__name__)


class BinanceMarketService(MarketService, BinanceApiSecretKeySpecifierMixin):
    market = BINANCE
    exchange = None
    config = None

    def __init__(
        self, config=None, api_key: str = None, secret_key: str = None
    ):
        super().__init__()

        if api_key is not None:
            self.binance_api_key = api_key

        if secret_key is not None:
            self.binance_secret_key = secret_key

        self.config = config

    def initialize(self, config):
        self.config = config

    def initialize_exchange(self, credentials = False):

        if credentials:
            if self.binance_api_key is None and self.binance_secret_key is None:
                self.binance_api_key = self.config\
                    .get(BINANCE_API_KEY, None)
                self.binance_secret_key = \
                    self.config.get(BINANCE_SECRET_KEY, None)

            exchange_class = getattr(ccxt, BINANCE_CCXT_ID)
            self.exchange = exchange_class({
                'apiKey': self.get_api_key(),
                'secret': self.get_secret_key(),
            })
        else:
            exchange_class = getattr(ccxt, BINANCE_CCXT_ID)
            self.exchange = exchange_class({})

    def pair_exists(self, target_symbol: str, trading_symbol: str):
        self.initialize_exchange()

        try:
            data = self.get_ticker(target_symbol, trading_symbol)
            return "symbol" in data
        except OperationalException:
            return False

    def get_ticker(self, target_symbol: str, trading_symbol: str):
        self.initialize_exchange()

        try:
            symbol = f"{target_symbol.upper()}{trading_symbol.upper()}"
            return self.exchange.fetchTicker(symbol)
        except Exception as e:
            logger.exception(e)
            raise OperationalException(
                f"Could not retrieve ticker"
                f"{target_symbol.upper()}{trading_symbol.upper()}"
            )

    def get_order_book(self, target_symbol: str, trading_symbol: str):
        self.initialize_exchange()

        try:
            symbol = f"{target_symbol.upper()}/{trading_symbol.upper()}"
            return self.exchange.fetchOrderBook(symbol)
        except Exception as e:
            logger.exception(e)
            raise OperationalException(
                "Could not retrieve order book"
            )

    def get_balance(self, symbol: str = None):
        self.initialize_exchange(credentials=True)
 
        try:
            balances = self.exchange.fetch_balance()["info"]["balances"]
        except Exception as e:
            logger.exception(e)
            raise OperationalException(
                "Could not retrieve balance"
            )

        if symbol is not None:

            for balance in balances:

                if balance["asset"] == symbol.upper():
                    return balance['free']

            return None
        else:
            return balances

    def create_limit_buy_order(
            self,
            target_symbol: str,
            trading_symbol: str,
            amount: float,
            price: float
    ):
        self.initialize_exchange(credentials=True)

        symbol = f"{target_symbol.upper()}/{trading_symbol.upper()}"

        try:
            self.exchange.create_limit_buy_order(
                symbol, amount, price
            )
        except Exception as e:
            logger.exception(e)
            raise OperationalException("Could not create limit buy order")

    def create_limit_sell_order(
        self,
        target_symbol: str,
        trading_symbol: str,
        amount: float,
        price: float
    ):
        self.initialize_exchange(credentials=True)

        symbol = f"{target_symbol.upper()}/{trading_symbol.upper()}"

        try:
            self.exchange.create_limit_sell_order(
                symbol, amount, price
            )
        except Exception as e:
            logger.exception(e)
            raise OperationalException("Could not create limit sell order")

    def create_market_buy_order(
        self,
        target_symbol: str,
        trading_symbol: str,
        amount: float,
    ):
        self.initialize_exchange(credentials=True)

        symbol = f"{target_symbol.upper()}/{trading_symbol.upper()}"

        try:
            self.exchange.create_market_buy_order(symbol, amount)
        except Exception as e:
            logger.exception(e)
            raise OperationalException("Could not create market buy order")

    def create_market_sell_order(
        self,
        target_symbol: str,
        trading_symbol: str,
        amount: float,
    ):
        self.initialize_exchange(credentials=True)

        symbol = f"{target_symbol.upper()}/{trading_symbol.upper()}"

        try:
            self.exchange.create_market_buy_order(symbol, amount)
        except Exception as e:
            logger.exception(e)
            raise OperationalException("Could not create market sell order")

    def cancel_order(self, order):
        self.exchange.cancelOrder(
            order.get_order_reference(),
            f"{order.get_target_symbol()}/{order.get_trading_symbol()}")

    def get_orders(self, target_symbol: str, trading_symbol: str):
        self.initialize_exchange(credentials=True)

        try:
            symbol = f"{target_symbol.upper()}/{trading_symbol.upper()}"
            return self.exchange.fetchOrders(symbol)
        except Exception as e:
            logger.exception(e)
            raise OperationalException("Could not retrieve orders")

    def get_order(self, order_id, target_symbol: str, trading_symbol: str):
        self.initialize_exchange(credentials=True)

        try:
            symbol = f"{target_symbol.upper()}/{trading_symbol.upper()}"
            return self.exchange.fetchOrder(order_id, symbol)
        except Exception as e:
            logger.exception(e)
            raise OperationalException("Could not retrieve order")

    def get_open_orders(
        self, target_symbol: str = None, trading_symbol: str = None
    ):
        self.initialize_exchange(credentials=True)

        try:
            if target_symbol is None or trading_symbol is None:
                return self.exchange.fetchOpenOrders()

            symbol = f"{target_symbol.upper()}/{trading_symbol.upper()}"
            return self.exchange.fetchOpenOrders(symbol)
        except Exception as e:
            logger.exception(e)
            raise OperationalException("Could not retrieve open orders")

    def get_closed_orders(
        self, target_symbol: str = None, trading_symbol: str = None
    ):
        self.initialize_exchange(credentials=True)

        try:
            if target_symbol is None or trading_symbol is None:
                return self.exchange.fetchClosedOrders()

            symbol = f"{target_symbol.upper()}/{trading_symbol.upper()}"
            return self.exchange.fetchClosedOrders(symbol)
        except Exception as e:
            logger.exception(e)
            raise OperationalException("Could not retrieve closed orders")

    def get_prices(
        self,
        target_symbol: str,
        trading_symbol: str,
        time_interval: TimeInterval
    ):
        pass



from flask_apscheduler import APScheduler

from investing_algorithm_framework.core.exceptions import OperationalException
from investing_algorithm_framework.core.models import TimeUnit, db, \
    TradingDataTypes
from investing_algorithm_framework.utils import random_string


class Strategy:

    def __init__(
        self,
        decorated=None,
        worker_id=None,
        time_unit=None,
        interval=None,
        data_provider_identifier=None,
        trading_data_type=None,
        trading_data_types=None,
        target_symbol=None,
        target_symbols=None,
        trading_symbol=None,
    ):
        self.data_provider_identifier = data_provider_identifier
        self.trading_data_types = trading_data_types
        self.target_symbol = target_symbol
        self.target_symbols = target_symbols
        self.trading_symbol = trading_symbol
        self.trading_data_type = None
        self.trading_data_types = None

        if trading_data_type is not None:
            self.trading_data_type = TradingDataTypes\
                .from_value(trading_data_type)

        if trading_data_types is not None:
            types = [TradingDataTypes.from_value(trading_data_type)
                     for trading_data_type in trading_data_types]

            self.trading_data_types = types

        from investing_algorithm_framework.core.context import algorithm

        self.decorated = decorated
        self.worker_id = worker_id

        if self.worker_id is None:

            if self.decorated:
                self.worker_id = self.decorated.__name__

                if self.decorated.__name__ in algorithm.workers:
                    self.id = random_string(10)
            else:
                self.id = random_string(10)

        if time_unit is None:
            raise OperationalException(
                "Time unit for strategy is not specified"
            )

        if isinstance(time_unit, TimeUnit):
            self.time_unit = time_unit.value
        else:
            self.time_unit = TimeUnit.from_string(time_unit).value

        if interval is None:
            raise OperationalException(
                "Time interval for strategy is not specified"
            )

        self.interval = interval

        if self.target_symbol is None and self.target_symbols is None \
                and (self.trading_data_type is not None
                     or self.trading_data_types is not None):
            raise OperationalException(
                "No target symbol or target symbols specified "
                "for trading data type"
            )

        algorithm.add_strategy(self)

    def add_to_scheduler(self, app_scheduler: APScheduler):
        if TimeUnit.SECONDS.equals(self.time_unit):
            app_scheduler.add_job(
                id=self.worker_id,
                name=self.worker_id,
                func=self.__call__,
                trigger="interval",
                seconds=self.interval
            )
        elif TimeUnit.MINUTE.equals(self.time_unit):
            app_scheduler.add_job(
                id=self.worker_id,
                name=self.worker_id,
                func=self.__call__,
                trigger="interval",
                minutes=self.interval
            )
        elif TimeUnit.HOUR.equals(self.time_unit):
            app_scheduler.add_job(
                id=self.worker_id,
                name=self.worker_id,
                func=self.__call__,
                trigger="interval",
                minutes=(self.interval * 60)
            )

    def run_strategy(
        self,
        algorithm_context,
        ticker,
        tickers,
        order_book,
        order_books,
        **kwargs
    ):
        raise NotImplementedError("Run strategy method is not implemented")

    def __call__(
        self,
        time_unit: TimeUnit = TimeUnit.MINUTE.value,
        interval=10,
        data_provider_identifier=None,
        trading_data_type=None,
        trading_data_types=None,
        target_symbol=None,
        target_symbols=None,
        trading_symbol=None
    ):
        from investing_algorithm_framework import current_app as app
        data = {}

        if self.data_provider_identifier is not None:
            data = app.algorithm.get_data(
                data_provider_identifier=self.data_provider_identifier,
                trading_data_type=self.trading_data_type,
                trading_data_types=self.trading_data_types,
                target_symbols=self.target_symbols,
                target_symbol=self.target_symbol,
                trading_symbol=self.trading_symbol
            )

        # Run the decorated function in app context
        with db.app.app_context():

            if self.decorated:

                if data is not None:
                    self.decorated(app.algorithm, **data)
                else:
                    self.decorated(app.algorithm)
            else:
                self.run_strategy(app.algorithm, **data)



from investing_algorithm_framework.configuration.constants import BINANCE
from investing_algorithm_framework.core.order_executors.binance \
    import BinanceOrderExecutor


class DefaultOrderExecutorFactory:

    @staticmethod
    def of_market(market):

        if market.upper() == BINANCE:
            return BinanceOrderExecutor()

        return None



from investing_algorithm_framework import current_app
from .operational_views import blueprint as operational_views_blueprint
from investing_algorithm_framework.views.order_views import blueprint \
    as order_views_blueprint
from investing_algorithm_framework.views.position_views import blueprint \
    as position_views_blueprint
from investing_algorithm_framework.views.portfolio_views import blueprint \
    as portfolio_views_blueprint

app = current_app
app.register_blueprint(operational_views_blueprint)
app.register_blueprint(order_views_blueprint)
app.register_blueprint(position_views_blueprint)
app.register_blueprint(portfolio_views_blueprint)



import functools
import threading


def synchronized(function):
    lock = threading.Lock()

    @functools.wraps(function)
    def wrapper(self, *args, **kwargs):
        with lock:
            return function(self, *args, **kwargs)
    return wrapper



import logging

from flask import Blueprint, jsonify

from investing_algorithm_framework.globals import current_app
from investing_algorithm_framework.exceptions import ApiException

logger = logging.getLogger(__name__)
blueprint = Blueprint("operational-views", __name__)


@blueprint.route("/start", methods=["GET"])
def start_algorithm():
    """
    View for starting the algorithm. If the algorithm is already
    running an exception will be thrown. Otherwise all workers
    will be started and the algorithm will be running.
    """

    if current_app.algorithm.running:
        raise ApiException("Algorithm is already running")

    current_app.algorithm.start()
    return jsonify({"message": "Algorithm started"}), 200


@blueprint.route("/stop", methods=["GET"])
def stop_algorithm():
    """
    View for stopping the algorithm. If the algorithm is already
    stopped an exception will be thrown. Otherwise all workers
    will be stopped.
    """

    if not current_app.algorithm.running:
        raise ApiException("Algorithm is already stopped")

    current_app.algorithm.stop()
    return jsonify({"message": "Algorithm stopped"}), 200



from abc import abstractmethod
from datetime import datetime as dt, timedelta as td
from investing_algorithm_framework.core.models import db, TimeFrame, \
    TimeInterval
from investing_algorithm_framework.core.models.model_extension \
    import SQLAlchemyModelExtension


class AssetPriceHistory:

    @abstractmethod
    def get_prices(self):
        pass

    @abstractmethod
    def get_market(self):
        pass

    @abstractmethod
    def get_trading_symbol(self):
        pass

    @abstractmethod
    def get_target_symbol(self):
        pass

    @abstractmethod
    def get_time_frame(self):
        pass

    @abstractmethod
    def get_updated_at(self):
        pass


class SQLLiteAssetPriceHistory(db.Model, SQLAlchemyModelExtension):
    __tablename__ = "asset_price_histories"

    id = db.Column(db.Integer, primary_key=True)
    target_symbol = db.Column(db.String, nullable=False)
    trading_symbol = db.Column(db.String, nullable=False)
    market = db.Column(db.String, nullable=False)
    time_frame = db.Column(db.String, nullable=False)
    updated_at = db.Column(db.DateTime, nullable=False)

    # Relationships
    prices = db.relationship(
        "SQLLiteAssetPrice",
        back_populates="asset_price_history",
        lazy="dynamic",
        cascade="all,delete",
    )

    def __init__(self, market, target_symbol, trading_symbol, time_frame):
        self.market = market
        self.target_symbol = target_symbol
        self.trading_symbol = trading_symbol
        self.time_frame = TimeFrame.from_value(time_frame).value

    @staticmethod
    def of(market: str, target_symbol: str, trading_symbol: str, time_frame):
        asset_price_history = SQLLiteAssetPriceHistory.query.filter_by(
            market=market,
            target_symbol=target_symbol,
            trading_symbol=trading_symbol,
            time_frame=TimeFrame.from_value(time_frame).value
        ).first()

        if not asset_price_history:
            asset_price_history = SQLLiteAssetPriceHistory(
                market=market,
                target_symbol=target_symbol,
                trading_symbol=trading_symbol,
                time_frame=TimeFrame.from_value(time_frame).value
            )
            asset_price_history.get_prices()
            asset_price_history.save(db)

        return asset_price_history

    def get_prices(self):
        time_frame = TimeFrame.from_value(self.time_frame)

        # Get the latest price point
        if TimeInterval.CURRENT.equals(time_frame.time_interval):
            prices = self._get_prices_current()

        # Get the latest minute prices
        elif TimeInterval.MINUTES_ONE.equals(time_frame.time_interval):
            prices = self._get_prices_one_minute()

        # Get the latest 15 minute prices
        elif TimeInterval.MINUTES_FIFTEEN.equals(time_frame.time_interval):
            prices = self._get_prices_fifteen_minutes()

        # Get the latest hourly prices
        elif TimeInterval.HOURS_ONE.equals(time_frame.time_interval):
            prices = self._get_prices_one_hour()

        # Get the latest four hour prices
        elif TimeInterval.HOURS_FOUR.equals(time_frame.time_interval):
            prices = self._get_prices_four_hour()

        # Get the latest daily prices
        elif TimeInterval.DAYS_ONE.equals(time_frame.time_interval):
            prices = self._get_prices_one_day()

        return prices.all()

    def _get_prices_current(self):
        from investing_algorithm_framework import current_app as app

        if self.updated_at is None \
                or self.updated_at < dt.utcnow() - td(seconds=5) \
                or not self.prices:
            self.remove_prices()

            market_service = app.algorithm.get_market_service(self.market)

            asset_prices = market_service.get_price(
                self.target_symbol,
                self.trading_symbol,
            )

            self.updated_at = dt.utcnow()
            self.prices = asset_prices
        return self.prices

    def _get_prices_one_minute(self):
        from investing_algorithm_framework import current_app as app
        market_service = app.algorithm.get_market_service(self.market)

        if self.updated_at is None\
                or self.updated_at < dt.utcnow() - td(minutes=1) \
                or not self.prices:
            self.remove_prices()

            asset_prices = market_service.get_prices(
                self.target_symbol,
                self.trading_symbol,
                TimeInterval.MINUTES_ONE,
            )

            self.updated_at = dt.utcnow()
            self.prices = asset_prices
        return self.prices

    def _get_prices_fifteen_minutes(self):
        from investing_algorithm_framework import current_app as app
        market_service = app.algorithm.get_market_service(self.market)

        if self.updated_at is None \
                or self.updated_at < dt.utcnow() - td(minutes=15) \
                or not self.prices:
            self.remove_prices()

            asset_prices = market_service.get_prices(
                self.target_symbol,
                self.trading_symbol,
                TimeInterval.MINUTES_FIFTEEN,
            )

            self.updated_at = dt.utcnow()
            self.prices = asset_prices
        return self.prices

    def _get_prices_one_hour(self):
        from investing_algorithm_framework import current_app as app
        market_service = app.algorithm.get_market_service(self.market)

        if self.updated_at is None \
                or self.updated_at < dt.utcnow() - td(hours=1) \
                or not self.prices:
            self.remove_prices()

            asset_prices = market_service.get_prices(
                self.target_symbol,
                self.trading_symbol,
                TimeInterval.HOURS_ONE,
            )
            self.updated_at = dt.utcnow()
            self.prices = asset_prices
        return self.prices

    def _get_prices_four_hour(self):
        from investing_algorithm_framework import current_app as app
        market_service = app.algorithm.get_market_service(self.market)

        if self.updated_at is None \
                or self.updated_at < dt.utcnow() - td(hours=4) \
                or not self.prices:
            self.remove_prices()

            asset_prices = market_service.get_prices(
                self.target_symbol,
                self.trading_symbol,
                TimeInterval.HOURS_FOUR
            )
            self.updated_at = dt.utcnow()
            self.prices = asset_prices
        return self.prices

    def _get_prices_one_day(self):
        from investing_algorithm_framework import current_app as app
        market_service = app.algorithm.get_market_service(self.market)

        if self.updated_at is None \
                or self.updated_at < dt.utcnow() - td(days=1) \
                or not self.prices:
            self.remove_prices()

            asset_prices = market_service.get_prices(
                self.target_symbol,
                self.trading_symbol,
                TimeInterval.DAYS_ONE
            )
            self.updated_at = dt.utcnow()
            self.prices = asset_prices

        return self.prices

    def _get_prices_one_week(self):
        from investing_algorithm_framework import current_app as app
        market_service = app.algorithm.get_market_service(self.market)

        if self.updated_at is None \
                or self.updated_at < dt.utcnow() - td(weeks=1) \
                or not self.prices:
            self.remove_prices()

            asset_prices = market_service.get_prices(
                self.target_symbol,
                self.trading_symbol,
                TimeInterval.WEEKS_ONE,
            )
            self.updated_at = dt.utcnow()
            self.prices = asset_prices

        return self.prices

    def remove_prices(self):
        from investing_algorithm_framework.core.models.snapshots import \
            SQLLiteAssetPrice

        db.session.query(SQLLiteAssetPrice).filter(
            SQLLiteAssetPrice.asset_price_history_id == self.id
        ).delete()
